	#!/usr/bin/python
# -*- coding: utf-8 -*-

#Autor: Erickson Silva 
#Email: <erickson.silva@lavid.ufpb.br> <ericksonsilva@live.com>

#LAViD - Laboratório de Aplicações de Vídeo Digital

import alexp
from AplicaRegras import *
import logging
import traceback

class TraduzSentencas(object):
	'''Realiza a tradução do texto em português para glosa
	'''
	
	def __init__(self):
		'''Instancia os aplicadores de regras e sinônimos.
		'''
		self.aplic_regras = AplicaRegras()
		#logging.basicConfig(filename='translate.log',
        #                    format='%(asctime)s - %(levelname)s:\n\n%(message)s\n\n\n##############################################\n\n',
        #                    level=logging.ERROR)

	def iniciar_traducao(self, sentenca, taxa=False):
		'''Metódo responsável por executar todos componentes necessários para a geração da glosa.
		'''
		try:
			has_sintatica = True
			analise_sintatica = alexp.run(sentenca)
		except Exception as ex:
			#self.salvar_log(str(traceback.format_exc()))
			analise_sintatica = None
			has_sintatica = False

		analise_morfologica = alexp.getAnaliseMorfologica()

		if (isinstance(analise_sintatica,type(None))):
			regras_aplicadas = self.aplic_regras.aplicar_regras_morfo(analise_morfologica)
		else:
			try:
				regras_aplicadas = self.aplic_regras.aplicar_regras_sint(analise_morfologica, analise_sintatica)
			except:
				regras_aplicadas = self.aplic_regras.aplicar_regras_morfo(analise_morfologica)

		sentenca_corrigida = self.aplic_regras.simplificar_sentenca(regras_aplicadas)
		glosa = " ".join([x[0] for x in sentenca_corrigida])
		glosa_sem_acentos = self.aplic_regras.remover_acento(glosa).upper().encode('utf-8')

		if glosa_sem_acentos:
			if taxa:
				taxa_qualidade = self.gerar_metrica_qualidade(sentenca_corrigida)
				return {'glosa':glosa_sem_acentos, 'taxa':taxa_qualidade, 'sintatica':has_sintatica}
			return glosa_sem_acentos
		return "TEXTO ERRADO ESCOLHER OUTRO"		

	def salvar_log(self, erro):
		'''Salva traceback de uma excessão do analisador sintático
		'''
		logging.error(erro)

	def gerar_metrica_qualidade(self, lista):
		#TODO: resolver path do arquivo
		arqSinais = open("sinais.txt", "r").read().split()
		quantSinaisTotal = len(lista)
		quantSinaisEncontradas = 0
		for x in lista:
			if x[0].upper()+".anim" in arqSinais:
				quantSinaisEncontradas += 1
			else:
				if x[1] == "NPR":
					quantSinaisTotal-=1
		return float(quantSinaisEncontradas)/quantSinaisTotal
